
/**
 * Implementation of hash table
 */

public class HashTable implements Dictionary_adt{
  /**
   * Costructs the hash table.
   */

private static final int DEFAULT_TABLE_SIZE = 23;

/** The array of elements. */
protected Comparable [ ] table;   // The array of elements
protected boolean [] isActive;
protected int currentSize;       // The number of occupied cells
protected boolean isRehashable;  //true if table can be expanded
protected int k =3;              //Coefficient for LinearProbing
private int numeroCollisioni;

  public HashTable() {
    this(DEFAULT_TABLE_SIZE);
  }


  /**
   * Costructs the hash table.
   * @param size the initial size of the hash table.
   */
  public HashTable(int size) {
   this(size,false);
 }


 /**
   * Costructs the hash table.
   * @param size the initial size of the hash table.
   * @param rehash true if the table can be expanded, false otherwise.
   */
  public HashTable(int size, boolean rehash) {
    if (!isPrime(size)){
       int tz = nextPrime(size);
       allocateTable(tz);
     }else{
       allocateTable(size);
     }
   makeEmpty();
   isRehashable = rehash;
 }



  /**
   * Inserts the key x in this dictionary.
   * @param key the element to insert
   * @return the key if the table has at least one empty location, null otherwise
   */
  public Object insert(Comparable key ){
    int collisionNum = 0;
    int initialPos = hash( key.toString(),table.length );
    int currentPos=initialPos;
    int insertPos= -1;

    while(collisionNum<table.length-1){
      if (table[currentPos] == null){
        if (insertPos == -1)
          insertPos = currentPos;
        if (!isActive[currentPos])
          break;
      } else if (table[ currentPos ].compareTo( key )==0) { //table[currentPos]!= null
        	System.out.println("Element "+ key +" is alredy in the hash table.");
             	return key;
      	} /* End else */
      currentPos = initialPos + k * ++collisionNum;  // Compute kth probe
      currentPos = currentPos % table.length; // Implement the mod
    }

    if (insertPos!= -1){
      table[ insertPos ] = key;
      isActive[ insertPos] = true;
      ++currentSize;
      System.out.println("Insertion of "+key+": in position " +currentPos);
      if((isRehashable)&&(currentSize > table.length/2)){
        System.out.println("Rehash!");
        rehash();
      }
      return key;
    }else {
      System.out.println("Insertion impossible: hash table full");
      return null;
    }
  }

  /**
   * Removes the mapping for this key from this dictionary if it is present
   * @param key the element to remove.
   */
  public Comparable remove(Comparable key ){
    int collisionNum = 0;
    int initialPos = hash( key.toString(),table.length );
    int currentPos=initialPos;

    System.out.println("initial Position of "+key+"is "+initialPos);
    System.out.print("current Position of "+key+"is "+currentPos);

    while ( ( ((table[currentPos] == null)&& isActive[currentPos] )||
    ( (table[currentPos] != null) && (table[currentPos].compareTo(key)!=0))) &&
             (collisionNum < table.length-1)) {
      currentPos = initialPos + k * ++collisionNum; // Compute ith probe
      currentPos = currentPos % table.length; // Implement the mod
      System.out.print(", "+currentPos);
    }

    if ((table[currentPos]==null)||table[ currentPos ].compareTo( key )!=0){
      System.out.println("Element "+ key +" isn't in the hash table.");
      return null;
    }else{
      System.out.println("Element "+ key +" is in the hash table.");
      table[currentPos]=null;
      return key;
    }
  }

  /**
   * Returns the value to which this map maps the specified key.
   * @param key key whose associated value is to be returned.
   * @return the value to which this map maps the specified key,
   * or null if the map contains no mapping for this key.
   */
 public Object find (Comparable key ) {
   int collisionNum = 0;
   int initialPos = hash( key.toString(),table.length );
   int currentPos=initialPos;

    System.out.println("initial Position of "+key+"is "+initialPos);
    System.out.print("current Position of "+key+"is "+currentPos);

      while ( ( ((table[currentPos] == null) && isActive[currentPos] )||
               ( (table[currentPos] != null) && (table[currentPos].compareTo(key)!=0))) &&
             (collisionNum < table.length-1)) {
        currentPos = initialPos + k * ++collisionNum; // Compute ith probe
        currentPos = currentPos % table.length; // Implement the mod
        System.out.print(", "+currentPos);
      }


    if ((table[currentPos]==null)||table[ currentPos ].compareTo( key )!=0){
      System.out.println("Element "+ key +" isn't in the hash table.");
      return null;
    }else{
      System.out.println("Element "+ key +" is in the hash table.");
      return key;
    }

 }


 /**
 * A hash routine for String objects.
 * @param key the String to hash.
 * @param tableSize the size of the hash table.
 * @return the hash value.
 */
 public static int hash( String key, int tableSize ){
   int hashVal = 0;

   for( int i = 0; i < key.length( ); i++ )
     hashVal = 37 * hashVal + key.charAt( i );

   hashVal %= tableSize;
   if( hashVal < 0 )
     hashVal += tableSize;

    return hashVal;
  }



 /**
 * Make the hash table logically empty.
 */
 public void makeEmpty( ){
   currentSize = 0;
   for( int i = 0; i < table.length; i++ ){
     table[i] = null;
     isActive[i] = false;
   }
  }


  /**
  * Internal method to allocate array.
  * @param TSize the size of the table.
  */
  private void allocateTable( int TSize ){

      table = new Comparable[TSize];
      isActive =new boolean [TSize];
      System.out.println("Table Length: "+ TSize);
  }


  /**
  * Internal method to find a prime number at least as large as n.
  * @param n the starting number (must be positive).
  * @return a prime number larger than or equal to n.
  */
  private static int nextPrime( int n ){
    if( n % 2 == 0 )
      n++;

    for( ; !isPrime( n ); n += 2 )
      ;

    return n;
  }


  /**
  * Internal method to test if a number is prime.
  * Not an efficient algorithm.
  * @param n the number to test.
  * @return the result of the test.
  */
  private static boolean isPrime( int n ){
    if( n == 2 || n == 3 )
      return true;

    if( n == 1 || n % 2 == 0 )
      return false;

    for( int i = 3; i * i <= n; i += 2 )
      if( n % i == 0 )
        return false;

    return true;
  }


  /**
   * Expand the hash Table
   */
  private void rehash( ){
    Comparable [ ] oldTable = table;

    // Create a new double-sized, empty table
    allocateTable( nextPrime( 2 * oldTable.length ) );
    makeEmpty();

    // Copy table over
    for( int i = 0; i < oldTable.length; i++ )
                if( oldTable[ i ] != null )
                    insert( oldTable[ i ] );

            return;
    }

	public void startCounter() {
		/* Inizializzazione contatore */
	}

	public int getColl() {
	int noColl = 0;
		/* Restituisce il no. di collisioni 
		registrate dall'avvio del contatore */
		return noColl;
	}

	public int maxClusterSize() {
	int maxClusterSize = 0;
		/* Calcola il no. di elementi presenti
		nel cluster piu' grande */
		return maxClusterSize;
	}

}